import os
import importlib

def node(f = None,tag = None,**forced_nodes):
    def decorator(g):
        g.node_tag = tag
        if len(forced_nodes) > 0:
            g.forced_nodes = forced_nodes
        return g
    
    if callable(f):
        f.node_tag = tag
        if len(forced_nodes) > 0:
            f.forced_nodes = forced_nodes
        return f
    else:
        return decorator

def dynamic_import(module_name, module_path):
    module_spec = importlib.util.spec_from_file_location(module_name, module_path)
    module = importlib.util.module_from_spec(module_spec)
    module_spec.loader.exec_module(module)
    return module

def dynamic_import_from_dir(module_dir):
    imported_dict = {}
    for root, dirs, files in os.walk(module_dir):
        for f in files:
            if f.endswith(".py"):
                module_path = os.path.join(root,f)
                module_name = f.split(".")[0]
                imported_module = dynamic_import(module_name, module_path)
                imported_dict.update(imported_module.__dict__)
    return imported_dict

def load_nodes(module_dir):
    imported_dict = dynamic_import_from_dir(module_dir)
    nodes = {k:v for k,v in imported_dict.items() if hasattr(v,"node_tag")}
    return nodes




from typing import Dict,List,Callable
import networkx as nx
from .graph_functions import nodes_graph,model_graph,graph_subcomponent_nodes
from .model_node import ModelNode

class Model():
    nodes: Dict[str,Callable]
    nodes_graph: nx.DiGraph
    nodes_with_forced_nodes: Dict[str,Dict]
    graph: nx.DiGraph
    inputs: List[str]
    call_order: List[str]
    model_nodes: Dict
    auxiliary_nodes:List[str]

    def __init__(self,nodes:Dict[str,Callable]):
        self.nodes = nodes
        self.nodes_graph = nodes_graph(nodes)
        self.nodes_with_forced_nodes = {k:(v.forced_nodes) for k,v in self.nodes.items() if hasattr(v,"forced_nodes")}
        self.graph = model_graph(self.nodes_graph,self.nodes_with_forced_nodes)
        self.inputs = list(set(self.nodes_graph.nodes()).difference(nodes.keys()))
        self.call_order = [node for node in list(nx.topological_sort(self.graph)) if node not in self.inputs]
        self.model_nodes = {k:ModelNode(k,self.nodes,self.nodes_with_forced_nodes,self.graph) for k in self.call_order}
        self.auxiliary_nodes = list(set(self.graph.nodes()).difference(self.nodes_graph.nodes()))

    def compute(self,input:Dict,keep_auxiliary_nodes:bool=False)->Dict:
        for node_name in self.call_order:
            model_node = self.model_nodes[node_name]
            call_input = [input[k] for k in model_node.inputs]
            input[node_name] = model_node.compute(*call_input)
        if not keep_auxiliary_nodes:
            for auxiliary_node in self.auxiliary_nodes:
                del input[auxiliary_node]
        return input
    
    def submodel(self,nodes_names:List[str]):
        subcomponent_nodes_names = graph_subcomponent_nodes(self.nodes_graph,nodes_names)
        submodel_nodes = {node_name:node for node_name,node in self.nodes.items() if node_name in subcomponent_nodes_names}
        submodel = Model(submodel_nodes)
        return submodel
    






from nodemodel.model import Model

def test_submodel():
    def z(a):
        return a

    def y(b):
        return b

    def x(c):
        return c

    def a(b):
        return b

    def b():
        return 1

    def c():
        return 1

    m = Model({"z":z,"y":y,"x":x,"a":a,"b":b,"c":c})
    assert m.compute({}) == {'b': 1, 'c': 1, 'y': 1, 'a': 1, 'x': 1, 'z': 1}
    m_sub = m.submodel(["z","y"])
    assert m_sub.compute({}) == {'b': 1, 'y': 1, 'a': 1, 'z': 1}



from nodemodel.model import Model

def test_model_1():
    def e(b):
        return b*5

    def c(b):
        return b
    c.forced_nodes = {"y":3}

    def b(a,y):
        return a + y
    b.forced_nodes = {"x":2}

    def a(x):
        return x

    nodes = {"a":a,"b":b,"c":c,"e":e}
    input = {"x":1,"y":1}

    m = Model(nodes)
    assert set(m.nodes_graph.edges()) == {('b', 'c'), ('a', 'b'), ('x', 'a'), ('b', 'e'), ('y', 'b')}
    assert set(m.graph.edges()) == {(('y', 3), ('b', 'y', 3)), 
                                    (('a', 'x', 2), ('b', 'y', 3)), 
                                    (('b', 'y', 3), 'c'), ('x', 'a'), 
                                    (('a', 'x', 2), 'b'), 
                                    (('x', 2), ('a', 'x', 2)), 
                                    ('b', 'e'), 
                                    ('y', 'b')}
    assert m.call_order == [('x', 2), ('y', 3), 'a', ('a', 'x', 2), 'b', ('b', 'y', 3), 'e', 'c']
    assert m.model_nodes[('x', 2)].compute() == 2
    assert m.model_nodes['b'].inputs == [('a', 'x', 2), 'y']
    assert m.model_nodes['c'].inputs == [('b', 'y', 3)]
    assert set(m.auxiliary_nodes) == {('b', 'y', 3), ('a', 'x', 2), ('x', 2), ('y', 3)}
    assert m.compute(input) == {'x': 1, 'y': 1, 'a': 1, 'b': 3, 'e': 15, 'c': 5}
    assert m.compute(input,keep_auxiliary_nodes=True) == {'x': 1, 'y': 1, ('x', 2): 2, ('y', 3): 3, 'a': 1, ('a', 'x', 2): 2, 
                                'b': 3, ('b', 'y', 3): 5, 'e': 15, 'c': 5}

def test_model_2():
    def c(a):
        return a
    c.forced_nodes = {"x":7}

    def b(a):
        return a
    b.forced_nodes = {"x":5}

    def a(x):
        return x
    nodes = {"a":a,"b":b,"c":c}
    inputs = {"x":1}

    m = Model(nodes)
    assert set(m.graph.edges()) == {('x', 'a'), 
                            (('x', 5), ('a', 'x', 5)), 
                            (('a', 'x', 7), 'c'),
                            (('x', 7), ('a', 'x', 7)), 
                            (('a', 'x', 5), 'b')}
    assert set(m.auxiliary_nodes) == {('x', 5), ('a', 'x', 7), ('x', 7), ('a', 'x', 5)}
    assert m.compute(inputs) == {'x': 1, 'a': 1, 'b': 5, 'c': 7}

def test_model_3():
    def c(a):
        return a
    c.forced_nodes = {"x":5}

    def b(a):
        return a
    b.forced_nodes = {"x":5}

    def a(x):
        return x
    nodes = {"a":a,"b":b,"c":c}
    inputs = {"x":1}

    m = Model(nodes)
    assert set(m.auxiliary_nodes) == {('x', 5), ('a', 'x', 5)}
    assert m.compute(inputs) == {'x': 1, 'a': 1, 'b': 5, 'c': 5}

def test_model_4():
    def a(x):
        return x
    a.forced_nodes = {"z":5}

    nodes = {"a":a}
    inputs = {"x":1}

    m = Model(nodes)
    assert set(m.auxiliary_nodes) == set()
    assert m.compute(inputs) == {'x': 1, 'a': 1}

def test_model_5():
    def b(a):
        return a
    b.forced_nodes = {"x":5,"y":3}

    def c(a):
        return a
    c.forced_nodes = {"y":3,"x":5}

    def a(x,y):
        return x + y
    nodes = {"a":a,"b":b,"c":c}
    inputs = {"x":1,"y":1}

    m = Model(nodes)
    assert set(m.auxiliary_nodes) == {('y', 3), ('x', 5), ('a', 'x', 5, 'y', 3)}
    assert m.nodes_with_forced_nodes == {'b': {'x': 5, 'y': 3}, 'c': {'y': 3, 'x': 5}}
    assert m.compute(inputs) == {'x': 1, 'y': 1, 'a': 2, 'b': 8, 'c': 8}
    assert m.compute(inputs,keep_auxiliary_nodes=True) == {'x': 1, 'y': 1, 'a': 2, 'b': 8, 'c': 8, 
                                                        ('x', 5): 5, ('y', 3): 3, ('a', 'x', 5, 'y', 3): 8}
    

def test_model_6():
    def a(x):
        return x
    a.forced_nodes = {"a":5}

    m = Model({"a":a})
    assert m.compute({"x":1}) == {"x":1,"a":1}


from nodemodel.utils import load_nodes
from nodemodel.model import Model
import os

node_1_code = '''from nodemodel import node

@node(x=2)
def b(a, y):
    return a + y

@node
def a(x):
    return x

def z(u):
    return u
'''

node_2_code = '''from nodemodel import node

@node
def e(b):
    return b * 5

@node(y=3)
def c(b):
    return b
'''

def create_folder_structure(tmp_path):
    submodule_path = os.path.join(tmp_path, "submodule")
    os.makedirs(submodule_path, exist_ok=True)

    init_path = os.path.join(tmp_path, "__init__.py")
    with open(init_path, 'w') as f:
        pass
    
    node_1_path = os.path.join(tmp_path, "node_1.py")
    with open(node_1_path, 'w') as f:
        f.write(node_1_code)

    init_submodule_path = os.path.join(submodule_path, "__init__.py")
    with open(init_submodule_path, 'w') as f:
        pass
    
    node_2_path = os.path.join(submodule_path, "node_2.py")
    with open(node_2_path, 'w') as f:
        f.write(node_2_code)


def test_load_nodes(tmp_path):
    create_folder_structure(tmp_path)
    nodes = load_nodes(tmp_path)
    m = Model(nodes)
    assert m.compute({"x":1,"y":1}) == {'x': 1, 'y': 1, 'a': 1, 'b': 3, 'e': 15, 'c': 5}



import networkx as nx
from typing import Dict,List,Callable
from .graph_functions import func_args

class ModelNode():
    compute:Callable
    inputs: List

    def __init__(self,node_name,nodes:Dict,nodes_with_forced_nodes:Dict,graph:nx.DiGraph):
        if node_name in nodes_with_forced_nodes.keys():
            self.compute = nodes[node_name]
            origin_inputs = func_args(self.compute)
            inputs_unordered = list(graph.predecessors(node_name))
            inputs_dict = {(k[0] if isinstance(k,tuple) else k):k for k in inputs_unordered}
            self.inputs = [inputs_dict[k] for k in origin_inputs]
        elif isinstance(node_name,str):
            self.compute = nodes[node_name]
            self.inputs = func_args(self.compute)
        elif isinstance(node_name,tuple) and len(node_name) == 2:
            self.compute = lambda x = node_name[1]: x
            self.inputs = []
        elif isinstance(node_name,tuple) and len(node_name) > 2:
            origin_node_name = node_name[0]
            self.compute = nodes[origin_node_name]
            origin_inputs = func_args(self.compute)
            inputs_unordered = list(graph.predecessors(node_name))
            inputs_dict = {(k[0] if isinstance(k,tuple) else k):k for k in inputs_unordered}
            self.inputs = [inputs_dict[k] for k in origin_inputs]


from nodemodel.utils import node
from nodemodel.model import Model

def test_node_decorator_1():
    @node
    def test():
        '''test_doc'''
        return 1

    assert test.__doc__ == "test_doc"
    assert test.__name__ == "test"
    assert test() == 1
    assert hasattr(test,"node_tag")
    assert not hasattr(test,"forced_nodes")

def test_node_decorator_2():
    @node()
    def test():
        '''test_doc'''
        return 1

    assert test.__doc__ == "test_doc"
    assert test.__name__ == "test"
    assert test() == 1
    assert hasattr(test,"node_tag")
    assert not hasattr(test,"forced_nodes")

def test_node_decorator_3():
    @node(tag = "my_tag")
    def test():
        '''test_doc'''
        return 1

    assert test.__doc__ == "test_doc"
    assert test.__name__ == "test"
    assert test() == 1
    assert hasattr(test,"node_tag")
    assert test.node_tag == "my_tag"
    assert not hasattr(test,"forced_nodes")

def test_node_decorator_4():
    @node(tag = "my_tag",a = 5,b = "c")
    def test():
        '''test_doc'''
        return 1

    assert test.__doc__ == "test_doc"
    assert test.__name__ == "test"
    assert test() == 1
    assert hasattr(test,"node_tag")
    assert test.node_tag == "my_tag"
    assert hasattr(test,"forced_nodes")
    assert test.forced_nodes == {'a':5,'b':"c"}

def test_node_decorator_with_model():
    @node
    def e(b):
        return b*5
    
    @node(y = 3)
    def c(b):
        return b

    @node(x = 2)
    def b(a,y):
        return a + y

    @node
    def a(x):
        return x
    
    nodes = {"a":a,"b":b,"c":c,"e":e}
    input = {"x":1,"y":1}

    m = Model(nodes)
    assert m.compute(input) == {'x': 1, 'y': 1, 'a': 1, 'b': 3, 'e': 15, 'c': 5}


import networkx as nx
from typing import List,Dict,Callable

def func_args(f:Callable)->List[str]:
    return list(f.__code__.co_varnames[:f.__code__.co_argcount])

def nodes_graph(nodes:Dict[str,Callable])->nx.DiGraph:
    edges = []
    for node_name,node in nodes.items():
        deps = func_args(node)
        for dep in deps:
            edges.append((dep,node_name))
    g = nx.DiGraph()
    g.add_edges_from(edges)
    return g

def custom_tuple_concat(a, b):
    if not isinstance(a, tuple):
        a = (a,)
    if not isinstance(b, tuple):
        b = (b,)
    return a + b

def get_cond_nodes(graph:nx.DiGraph,nodes_with_forced_nodes:Dict)->List:
    ordered_nodes = list(nx.topological_sort(graph))
    cond_nodes = [node for node in ordered_nodes if node in nodes_with_forced_nodes.keys()]
    return cond_nodes


def model_graph(nodes_graph,nodes_with_forced_nodes):
    model_graph = nodes_graph.copy()
    #Sort forced nodes in nodes_with_forced_nodes -> Important to mutualize values like {"a":1,"b":2} and {"b":2,"a":1}
    nodes_with_forced_nodes = {k:dict(sorted(v.items())) for k,v in nodes_with_forced_nodes.items()}
    #Get cond_nodes -> list of of all cond nodes
    cond_nodes = get_cond_nodes(model_graph,nodes_with_forced_nodes)
    #For each node with forced nodes, do the following:
    for cond_node in cond_nodes:
        #Get cond_node_ancestors_graph -> graph of all ancestors of cond_node  + cond_node
        cond_node_ancestors = nx.ancestors(model_graph,cond_node)
        cond_node_ancestors.add(cond_node)
        cond_node_ancestors_graph = model_graph.subgraph(cond_node_ancestors).copy()#Copy subgraph to avoid some edge cases
        #Rename nodes using forced_nodes info of cond_node
        forced_nodes = nodes_with_forced_nodes[cond_node]
        for forced_node in forced_nodes.keys():
            forced_node_value = forced_nodes[forced_node]
            if forced_node in cond_node_ancestors_graph:
                forced_node_descendants = list(nx.descendants(cond_node_ancestors_graph,forced_node))
                old_names = forced_node_descendants + [forced_node]
                new_names = [custom_tuple_concat(x,(forced_node,forced_node_value)) for x in forced_node_descendants]
                new_names.append((forced_node,forced_node_value))
                name_mapping = dict(zip(old_names,new_names))
                if cond_node in name_mapping:#Do not rename cond_node
                    del name_mapping[cond_node]
                cond_node_ancestors_graph = nx.relabel_nodes(cond_node_ancestors_graph,name_mapping)
        #Remove cond_node edges with predecessors in model_graph 
        cond_node_predecessors = list(model_graph.predecessors(cond_node))
        for predecessor in cond_node_predecessors:
            model_graph.remove_edge(predecessor, cond_node)
        #Combine cond_node_ancestors_graph with initial graph
        model_graph = nx.compose(model_graph,cond_node_ancestors_graph)
    return model_graph


def graph_subcomponent_nodes(graph:nx.DiGraph,nodes_names:List)->List:
    subcomponent_nodes = set()
    for node_name in nodes_names:
        subcomponent_nodes.update(nx.ancestors(graph, node_name))
    subcomponent_nodes.update(nodes_names)
    return list(subcomponent_nodes)

